#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Module containing MPI application wrapping support"""
import json
import logging
import os
import signal
import subprocess
import sys

from dlg import utils, droputils
from dlg.apps.app_base import BarrierAppDROP
from dlg.named_port_utils import (
    DropParser,
    get_port_reader_function,
    replace_named_ports,
)
from dlg.exceptions import InvalidDropException
from ..meta import (
    dlg_enum_param,
)

logger = logging.getLogger(__name__)


##
# @brief MPI
# @details An application component using the Message Passing Interface (MPI)
# @par EAGLE_START
# @param category Mpi
# @param tag template
# @param command /String/ComponentParameter/NoPort/ReadWrite//False/False/The command to be executed
# @param args /String/ComponentParameter/NoPort/ReadWrite//False/False/Additional command line arguments to be added to the command line to be executed
# @param num_of_procs 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Number of processes used for this application
# @param use_wrapper False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If the command should be executed inside the existing MPI communicator set this to True
# @param input_redirection /String/ComponentParameter/NoPort/ReadWrite//False/False/The command line argument that specifies the input into this application
# @param output_redirection /String/ComponentParameter/NoPort/ReadWrite//False/False/The command line argument that specifies the output from this application
# @param paramValueSeparator " "/String/ComponentParameter/NoPort/ReadWrite//False/False/Separator character(s) between parameters on the command line
# @param argumentPrefix "--"/String/ComponentParameter/NoPort/ReadWrite//False/False/Prefix to each keyed argument on the command line
# @param dropclass dlg.apps.mpi.MPIApp/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/the allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @param input_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Input port parsing technique
# @param output_parser pickle/Select/ComponentParameter/NoPort/ReadWrite/raw,pickle,eval,npy,path,dataurl/False/False/Output port parsing technique
# @par EAGLE_END
class MPIApp(BarrierAppDROP):
    """
    An application drop representing an MPI job.

    This application needs to be launched from within an MPI environment,
    and therefore the hosting NM must be part of an MPI communicator.
    This application uses MPI_Comm_Spawn to fire up the requested MPI
    application, which must *not* be aware of it having a parent.
    This drop will gather the individual exit codes from the launched
    applications and transition to ERROR if any of them did not exit cleanly,
    or to FINISHED if all of them finished successfully.

    """

    input_parser: DropParser = dlg_enum_param(DropParser, "input_parser", DropParser.PICKLE)  # type: ignore

    def initialize(self, **kwargs):
        super(MPIApp, self).initialize(**kwargs)

        self._maxprocs = self._popArg(kwargs, "maxprocs", 1)
        self._use_wrapper = self._popArg(kwargs, "use_wrapper", False)
        self._args = self._popArg(kwargs, "args", "")
        self._applicationArgs = self._popArg(kwargs, "applicationArgs", {})
        self._argumentPrefix = self._popArg(kwargs, "argumentPrefix", "--")
        self._paramValueSeparator = self._popArg(kwargs, "paramValueSeparator", " ")
        self._inputRedirect = self._popArg(kwargs, "input_redirection", "")
        self._outputRedirect = self._popArg(kwargs, "output_redirection", "")

        self._command = self._popArg(kwargs, "command", None)
        if not self._command:
            raise InvalidDropException(
                self, "No command specified, cannot create MPIApp"
            )
        self._recompute_data = {}

    def run(self):
        from mpi4py import MPI

        cmd, args = self._command, self._args
        inputs = self._inputs
        outputs = self._outputs

        logger.debug("Parameters found: %s", json.dumps(self.parameters))
        logger.debug("MPI Inputs: %s; MPI Outputs: %s", inputs, outputs)
        # we only support passing a path for bash apps
        fsInputs = {uid: i for uid, i in inputs.items() if droputils.has_path(i)}
        fsOutputs = {uid: o for uid, o in outputs.items() if droputils.has_path(o)}
        dataURLInputs = {
            uid: i for uid, i in inputs.items() if not droputils.has_path(i)
        }
        dataURLOutputs = {
            uid: o for uid, o in outputs.items() if not droputils.has_path(o)
        }
        # deal with named ports
        inport_names = self.parameters["inputs"] if "inputs" in self.parameters else []
        outport_names = (
            self.parameters["outputs"] if "outputs" in self.parameters else []
        )
        reader = get_port_reader_function(self.input_parser)
        keyargs, pargs = replace_named_ports(
            inputs.items(),
            outputs.items(),
            inport_names,
            outport_names,
            self._applicationArgs,
            argumentPrefix=self._argumentPrefix,
            separator=self._paramValueSeparator,
            parser=reader,
        )
        argumentString = (
            f"{' '.join(map(str,pargs + keyargs))}"  # add kwargs to end of pargs
        )
        # complete command including all additional parameters and optional redirects
        if len(argumentString.strip()) > 0:
            # the _cmdLineArgs would very likely make the command line invalid
            cmd = f"{self._command} {argumentString} "
        else:
            cmd = f"{self._command} {argumentString} {args} "
        if self._outputRedirect:
            cmd = f"{cmd} > {self._outputRedirect}"
        if self._inputRedirect:
            cmd = f"cat {self._inputRedirect} > {cmd}"
        cmd = cmd.strip()

        app_uid = self.uid

        # Replace inputs/outputs in command line with paths or data URLs
        cmd = droputils.replace_path_placeholders(cmd, fsInputs, fsOutputs)

        cmd = droputils.replace_dataurl_placeholders(cmd, dataURLInputs, dataURLOutputs)

        # Pass down daliuge-specific information to the subprocesses as environment variables
        env = os.environ.copy()
        env.update({"DLG_UID": self._uid})
        if self._dlg_session_id:
            env.update({"DLG_SESSION_ID": self._dlg_session_id})

        env.update({"DLG_ROOT": utils.getDlgDir()})

        logger.info("Command after wrapping is: %s", cmd)

        if self._use_wrapper:
            # We spawn this very same module
            # When invoked as a program (see at the bottom) this module
            # will get the parent communicator, run the program we're giving in the
            # command line, and send back the exit code.
            # Likewise, we barrier on the children communicator, and thus
            # we wait until all children processes are completed
            args = ["-m", __name__, cmd]
            cmd = sys.executable

        errcodes = []

        # Spawn the new MPI communicator and wait until it finishes
        # (it sends the stdout, stderr and exit codes of the programs)
        logger.info(
            "Executing MPI app in new communicator with %d ranks and command: %s %s",
            self._maxprocs,
            cmd,
            args,
        )

        vendor, version = MPI.get_vendor()  # @UndefinedVariable
        info = MPI.Info.Create()  # @UndefinedVariable
        logger.debug(
            "MPI vendor is %s, version %s",
            vendor,
            ".".join([str(x) for x in version]),
        )  # @UndefinedVariable
        comm_children = MPI.COMM_SELF.Spawn(
            cmd,
            args=args,
            maxprocs=self._maxprocs,
            errcodes=errcodes,
            info=info,
        )  # @UndefinedVariable

        n_children = comm_children.Get_remote_size()
        logger.info("%d MPI children apps spawned, gathering exit data", n_children)

        if self._use_wrapper:
            children_data = comm_children.gather(
                ("", "", 0), root=MPI.ROOT
            )  # @UndefinedVariable
            exit_codes = [x[2] for x in children_data]
            logger.info("Exit codes gathered from children processes: %r", exit_codes)

            any_failed = False
            for rank, (stdout, stderr, code) in enumerate(children_data):
                self._recompute_data[str(rank)] = [
                    code,
                    str(stdout),
                    str(stderr),
                ]
                if code == 0:
                    continue
                any_failed = True
                logger.error(
                    "stdout/stderr follow for rank %d:\nSTDOUT\n======\n%s\n\nSTDERR\n======\n%s",
                    rank,
                    stdout,
                    stderr,
                )

            if any_failed:
                raise Exception("One or more MPI children didn't exit cleanly")
        else:
            comm_children.barrier()

    def generate_recompute_data(self):
        return self._recompute_data


# When we are called by the MPIApp
def module_as_main():
    # Get the parent communicator before anything else happens
    # This way we ensure the communicator is valid
    from mpi4py import MPI

    parent_comm = MPI.Comm.Get_parent()  # @UndefinedVariable

    def handle(signNo, stack_frame):
        parent_comm.gather(("", "Received signal %d" % (signNo,), -1), root=0)

    signal.signal(signal.SIGINT, handle)
    signal.signal(signal.SIGTERM, handle)
    signal.signal(signal.SIGABRT, handle)

    # argv[0] is the name of this module
    # argv[1:] is the actual command + args
    try:
        proc = subprocess.Popen(
            sys.argv[1:],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            close_fds=False,
        )
        stdout, stderr = proc.communicate()
        code = proc.returncode
    except Exception as e:
        stdout, stderr, code = "", str(e), -1

    # Gather the results in the spawner rank and good bye
    parent_comm.gather((stdout, stderr, code), root=0)


if __name__ == "__main__":
    module_as_main()



import ast
from enum import Enum
import logging
import collections
import dlg.droputils as droputils
import dlg.drop_loaders as drop_loaders
from typing import Tuple

from dlg.data.drops import DataDROP

logger = logging.getLogger(__name__)


class DropParser(Enum):
    RAW = "raw"
    PICKLE = "pickle"
    EVAL = "eval"
    NPY = "npy"
    DILL = "dill"
    # JSON = "json"
    PATH = "path"  # input only
    DATAURL = "dataurl"  # input only
    BINARY = "binary"
    UTF8 = "utf-8"


def serialize_kwargs(keyargs, prefix="--", separator=" "):
    kwargs = []
    for name, value in iter(keyargs.items()):
        if prefix == "--" and len(name) == 1:
            kwargs += [f"-{name} {value}"]
        else:
            kwargs += [f"{prefix.strip()}{name.strip()}{separator}{str(value).strip()}"]
    logger.debug("kwargs after serialization: %s", kwargs)
    return kwargs


def clean_applicationArgs(applicationArgs: dict) -> dict:
    """
    Removes arguments with None and False values, if not precious. This
    is in particular used for Bash and Docker app command lines, else
    we would have empty values for command line arguments.

    Args:
        applicationsArgs (dict): the complete set of arguments

    Returns:
        dict: a dictionary with the relevant arguments only.
    """
    cleanedArgs = {}
    if not isinstance(applicationArgs, dict):
        logger.info("applicationArgs are not passed as a dict. Ignored!")
    for name, vdict in applicationArgs.items():
        if vdict in [None, False, ""]:
            continue
        elif isinstance(vdict, bool):
            vdict = {"precious": False, "value": "", "positional": False}
        elif isinstance(vdict, dict):
            precious = vdict["precious"]
            if vdict["value"] in [None, False, ""] and not precious:
                continue
        cleanedArgs.update({name: vdict})
    logger.debug("After clean_applicationArgs: %s", cleanedArgs)
    return cleanedArgs


def serialize_applicationArgs(applicationArgs, prefix="--", separator=" "):
    """
    Unpacks the applicationArgs dictionary and returns two strings, one for
    positional arguments and one for kw arguments that can be used to construct
    the final command line.
    """
    applicationArgs = clean_applicationArgs(applicationArgs)
    pargs = []
    kwargs = {}
    for name, vdict in applicationArgs.items():
        value = vdict["value"]
        positional = vdict["positional"]
        if positional:
            pargs.append(str(value).strip())
        else:
            kwargs.update({name: value})
    skwargs = serialize_kwargs(kwargs, prefix=prefix, separator=separator)
    logger.info("Constructed command line arguments: %s %s", pargs, kwargs)
    return (pargs, skwargs)


def identify_named_ports(
    port_dict: dict,
    positionalArgs: list,
    positionalPortArgs: dict,
    keywordArgs: dict,
    check_len: int = 0,
    mode: str = "inputs",
    parser: callable = None,
    addPositionalToKeyword: bool = False,
) -> dict:
    """
    Checks port names for matches with arguments and returns mapped ports.

    Args:
        port_dict (dict): ports {uid:name,...}
        positionalArgs (list): available positional arguments (will be modified)
        positionalPortArgs (dict): mapped arguments (will be modified)
        keywordArgs (dict): keyword arguments
        check_len (int): number of of ports to be checked
        mode (str ["inputs"]): mode, used just for logging messages
        parser (function): parser function for this port
        addPositionalToKeyword (bool): Adds a positional argument to the keyword
            argument dictionary. This is useful when you have arguments
            that can be specified both positionally and as keywords. For example,
            in a Python function, you might have `func(a, b=2)`, where `a` can be
            passed positionally or as a keyword.

    Returns:
        keywordPortArgs: dict of keyword arguments from named ports

    Side effect:
        Modifies:
        - positionalArgs
        - positionalPortArgs
        - keywordArgs

    """
    # p_name = [p["name"] for p in port_dict]
    logger.debug(
        "Using named ports to remove %s from arguments port_dict: %s, check_len: %d)",
        mode,
        port_dict,
        check_len,
    )
    logger.debug("Checking against keyargs: %s", keywordArgs)
    keywordPortArgs = {}
    positionalArgs = list(positionalArgs)
    keys = list(port_dict.keys())
    logger.debug("Checking ports: %s against %s %s", keys, positionalArgs, keywordArgs)
    for i in range(check_len):
        try:
            key = port_dict[keys[i]]["name"]
            value = port_dict[keys[i]]["path"]
        except KeyError as e:
            logger.debug("portDict: %s does not have key: %s", port_dict, keys[i])
            raise KeyError("")
        if value is None:
            value = ""  # make sure we are passing NULL drop events
        if key in positionalArgs:
            encoding = DropParser(positionalPortArgs[key]["encoding"])
            parser = get_port_reader_function(encoding)
            if parser:
                logger.debug("Reading from port using %s", parser.__repr__())
                value = parser(port_dict[keys[i]]["drop"])
            positionalPortArgs[key]["value"] = value
            logger.debug("Using %s '%s' for parg %s", mode, value, key)
            positionalArgs.remove(key)
            # We have positional argument that is also a keyword
            if addPositionalToKeyword:
                keywordPortArgs.update({key: positionalPortArgs[key]})
        elif key in keywordArgs:
            encoding = DropParser(keywordArgs[key]["encoding"])
            parser = get_port_reader_function(encoding)
            if parser:
                logger.debug("Reading from port using %s", parser.__repr__())
                value = parser(port_dict[keys[i]]["drop"])
            # if not found in appArgs we don't put them into portargs either
            # pargsDict.update({key: value})
            keywordArgs[key]["value"] = value
            keywordPortArgs.update({key: keywordArgs[key]})
            logger.debug("Using %s of type %s for kwarg %s", mode, type(value), key)
            _ = keywordArgs.pop(key)  # remove from original arg list
        else:
            logger.debug(
                "No matching argument found for %s key %s, %s, %s",
                mode,
                key,
                keywordArgs,
                positionalArgs,
            )

    logger.debug("Returning kw mapped ports: %s", keywordPortArgs)
    return keywordPortArgs


def check_ports_dict(ports: list) -> bool:
    """
    Checks whether all ports in ports list are of type dict. This is
    for backwards compatibility.

    Args:
        ports (list):

    Returns:
        bool: True if all ports are dict, else False
    """
    # all returns true if list is empty!
    logger.debug("Ports list is: %s", ports)
    if len(ports) > 0:
        return all(isinstance(p, dict) for p in ports)
    else:
        return False


def replace_named_ports(
    iitems: dict,
    oitems: dict,
    inport_names: dict,
    outport_names: dict,
    appArgs: dict,
    argumentPrefix: str = "--",
    separator: str = " ",
    parser: callable = None,
) -> Tuple[str, str]:
    """
    Function attempts to identify CLI component arguments that match port names.

    Inputs:
        iitems: itemized input port dictionary
        oitems: itemized output port dictionary
        inport_names: dictionary of input port names (key: uid)
        outport_names: dictionary of output port names (key: uid)
        appArgs: dictionary of all arguments
        argumentPrefix: prefix for keyword arguments
        separator: character used between keyword and value
        parser: reader function for ports

    This method is focused on creating two 'sets' of arguments:
    - The arguments that are passed to the application (keyword and
    positional arguments)
    - The arguments that are passed to the application that are derived from
    the ports of the drop (keywordPort and positionalPort arguments).

    Returns:
        tuple of serialized keyword arguments and positional arguments
    """
    logger.debug(
        "iitems: %s; inport_names: %s; outport_names: %s",
        iitems,
        inport_names,
        outport_names,
    )
    inputs_dict = collections.OrderedDict()
    for uid, drop in iitems:
        inputs_dict[uid] = {
            "drop": drop,
            "path": drop.path if hasattr(drop, "path") else "",
        }

    outputs_dict = collections.OrderedDict()
    for uid, drop in oitems:
        outputs_dict[uid] = {
            "drop": drop,
            "path": drop.path if hasattr(drop, "path") else "",
        }

    positionalArgs = _get_args(appArgs, positional=True)
    keywordArgs = _get_args(appArgs, positional=False)
    # we will need an ordered dict for all positional arguments
    # thus we create it here and fill it with values
    positionalPortArgs = collections.OrderedDict(positionalArgs)

    logger.debug(
        "posargs: %s; keyargs: %s, %s",
        positionalArgs,
        keywordArgs,
        check_ports_dict(inport_names),
    )

    keywordPortArgs = {}
    # Update the argument dictionaries in-place based on the port names.
    # This needs to be done for both the input ports and output ports on the drop.
    _process_port(
        inport_names,
        inputs_dict,
        keywordPortArgs,
        positionalArgs,
        positionalPortArgs,
        keywordArgs,
        iitems,
        parser,
        "inputs",
    )

    _process_port(
        outport_names,
        outputs_dict,
        keywordPortArgs,
        positionalArgs,
        positionalPortArgs,
        keywordArgs,
        oitems,
        parser,
        "outputs",
    )

    logger.debug("Arguments from ports: %s, %s,", keywordPortArgs, positionalPortArgs)

    # Clean arguments for Docker and Bash applications
    appArgs = clean_applicationArgs(appArgs)
    positionalArgs = _get_args(appArgs, positional=True)
    keywordArgs = _get_args(appArgs, positional=False)

    # Extract values from dictionaries - "encoding" etc. are irrelevant
    appArgs = {arg: subdict["value"] for arg, subdict in appArgs.items()}
    positionalArgs = {arg: subdict["value"] for arg, subdict in positionalArgs.items()}
    keywordArgs = {arg: subdict["value"] for arg, subdict in keywordArgs.items()}
    keywordPortArgs = {
        arg: subdict["value"] for arg, subdict in keywordPortArgs.items()
    }

    # Construct the final keywordArguments and positionalPortArguments
    for k, v in keywordPortArgs.items():
        if v not in [None, ""]:
            keywordArgs.update({k: v})
    for k, v in positionalPortArgs.items():
        logger.debug("port posarg %s has value %s", k, v)
        if k == "input_redirection":
            v = f"cat {v} > "
        if k == "output_redirection":
            v = f"> {v}"
        if v not in [None, ""]:
            positionalArgs.update({k: v})

    keywordArgs = (
        serialize_kwargs(keywordArgs, prefix=argumentPrefix, separator=separator)
        if len(keywordArgs) > 0
        else [""]
    )
    pargs = list(positionalArgs.values())
    if not pargs or None in pargs:
        pargs = [""]

    logger.debug("After port replacement: pargs: %s; keyargs: %s", pargs, keywordArgs)
    return keywordArgs, pargs


def _process_port(
    port_names,
    ports,
    keywordPortArgs,
    positionalArgs,
    positionalPortArgs,
    keywordArgs,
    iitems,
    parser,
    mode,
):
    """
    For the set of port names, perform a backwards compatible update of the:

        - Keyword Arguments (application, and port name)
        - Positional Arguments (application, and port name)

    Note: This performs an IN-PLACE transformation of the dictionaries, through the
    identify_named_ports() method.
    """

    if check_ports_dict(port_names):
        for port in port_names:
            key = list(port.keys())[0]
            ports[key].update({"name": port[key]})
        keywordPortArgs.update(
            identify_named_ports(
                ports,
                positionalArgs,
                positionalPortArgs,
                keywordArgs,
                check_len=len(iitems),
                mode=mode,
                parser=parser,
            )
        )
    else:
        for i in range(min(len(iitems), len(positionalArgs))):
            keywordPortArgs.update({list(positionalArgs)[i]: list(iitems)[i][1]})


def _get_args(appArgs, positional=False):
    """
    Separate out the arguments dependening on if we want positional or keyword style
    """
    args = {
        arg: {
            "value": appArgs[arg]["value"],
            "encoding": appArgs[arg].get("encoding", "dill"),
        }
        for arg in appArgs
        if (appArgs[arg]["positional"] == positional)
    }

    argType = "Positional" if positional else "Keyword"
    logger.debug("%s arguments: %s", argType, args)
    return args


def get_port_reader_function(input_parser: DropParser):
    """
    Return the function used to read input from a named port
    """
    # Inputs are un-pickled and treated as the arguments of the function
    # Their order must be preserved, so we use an OrderedDict
    if input_parser is DropParser.PICKLE:
        # all_contents = lambda x: pickle.loads(droputils.allDropContents(x))
        reader = drop_loaders.load_pickle
    elif input_parser is DropParser.EVAL:

        def optionalEval(x):
            # Null and Empty Drops will return an empty byte string
            # which should propogate back to None
            content: str = droputils.allDropContents(x).decode("utf-8")
            return ast.literal_eval(content) if len(content) > 0 else None
        reader = optionalEval
    elif input_parser is DropParser.UTF8:
        def utf8decode(drop: "DataDROP"):
            """
            Decode utf8
            Not stored in drop_loaders to avoid cyclic imports
            """
            return droputils.allDropContents(drop).decode("utf-8")
        reader = utf8decode
    elif input_parser is DropParser.NPY:
        reader = drop_loaders.load_npy
    elif input_parser is DropParser.PATH:
        reader = lambda x: x.path
    elif input_parser is DropParser.DATAURL:
        reader = lambda x: x.dataurl
    elif input_parser is DropParser.DILL:
        reader = drop_loaders.load_dill
    elif input_parser is DropParser.BINARY:
        reader = drop_loaders.load_binary
    else:
        raise ValueError(input_parser.__repr__())
    return reader



# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '4.9.0'
git_version = '927717d1de5a41075133fb61fe82491572a55c27'
full_version = '4.9.0'
is_release = True

if not is_release:
    version = full_version


#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Unit/regression testing for the application of named_port_utils.py

Use examples derived from the following DROP classes: 
- pyfunc
- s3_drop
- bash_shell_app
- dockerapp 
- mpi
"""
import dill
import logging
import unittest
import pytest

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

# Note this test will only run with a full installation of DALiuGE.
pexpect = pytest.importorskip("dlg.dropmake")

import json
from pathlib import Path

import dlg.droputils as droputils
import dlg.graph_loader as graph_loader

from dlg.ddap_protocol import DROPStates
from dlg.dropmake import path_utils

class TestPortsEncoding(unittest.TestCase):
    """
    Given a dropspec, make sure the ports are loaded correctly. 
    """ 
    def _create_and_run_graph_spec_from_lgt(self, logical_graph_name: str):
        """
        Boilerplate graph running code which takes a Logical Graph Template (LGT) and 
        performs the translation and submission through the graph loader. 
        
        Returns drop leaf nodes on graph completion. 
        """
        spec = Path(path_utils.get_lg_fpath("drop_spec", logical_graph_name))
        with Path(spec).open('r', encoding="utf-8") as f: 
            appDropSpec = json.load(f)
        
        roots = graph_loader.createGraphFromDropSpecList(appDropSpec)
        # drops = [v for d,v in drops.items()]
        leafs = droputils.getLeafNodes(roots)  
        with  droputils.DROPWaiterCtx(self, leafs, timeout=3):
            for drop in roots: 
                fut = drop.async_execute()
                fut.result()
    
        return leafs


    def test_pyfunc_ports_encoding(self):
        """
        
        This test evaluates the use of per-port encoding using the test_ports.graph, 
        described below. 
        
        The CreateMultiA drop has inputs of 2,8 (non-default), which are replicated
        with the chosen encoding (numpy and pickle, respectively). These are passed to 
        CreateMultiB, which again replicates the output with selected encoding. These are
        passed to FileDrops, from which we can confirm the encoding has worked as 
        expected. 


                    ------> numpy(2) ---              ------> "numpy(2)"
                    |                   |             |
                    (npy)              (npy)        (UTF-8) 
                    |                   |             |
        <CreateMultiA(2,8)>                  <CreateMultiB> 
                    |                   |             |
                    (pickle)          (pickle)       (pickle)
                    |                   |             |
                    ------> pickle(8) ---             -----> pickle(8)
        """

        leafs = self._create_and_run_graph_spec_from_lgt("test_ports.graph")               
        for l in leafs:
            self.assertEqual(DROPStates.COMPLETED, l.status) 

        # Leaf Node 1 has been encoded first as numpy, second as UTF-8. 
        leaf = leafs.pop()
        desc = leaf.open()
        self.assertEqual(8, dill.loads(leaf.read(desc)))
        leaf = leafs.pop()
        desc = leaf.open()
        self.assertEqual("array(2)", leaf.read(desc).decode())
        

    @unittest.skip
    def test_bash_shell_ports(self): 
        """
        "drop_spec", "pyfunc_glob_shell_test.graph"
        """
        leafs = self._create_and_run_graph_spec_from_lgt("pyfunc_glob_shell_test.graph")
        
        for l in leafs:
            self.assertEqual(DROPStates.COMPLETED, l.status) 
        self.assertEqual(0, leafs.pop().proc.returncode)
